//----------------------------------------------------------------------------
// ObjectWindows - (C) Copyright 1991, 1993 by Borland International
//   Example program that shows Dibs, Bitmaps and Palettes in a scrolling
//   Window. Also uses diagnostics to trace thru some routines
//----------------------------------------------------------------------------
#include <owl\owlpch.h>
#include <owl\applicat.h>
#include <owl\dc.h>
#include <owl\menu.h>
#include <owl\framewin.h>
#include <owl\scroller.h>
#include <owl\opensave.h>
#include <owl\clipview.h>
#include <string.h>
#include <dir.h>
#include <stdio.h>
#include <mem.h>
#include "jpeglib.h"
#include "jpgview.h"

DIAG_DEFINE_GROUP_INIT(OWL_INI, BmpView, 1, 0);

#define USE_DDB     // Use a DDB, else blts the DIB directly to the screen
//#define PAINT_ICON  // define to paint the icon from the bitmap

#define MAXAPPNAME  20
static const char AppName[] = "BitmapView";

//
// TBmpViewWindow, a Bitmap displaying window derived from TClipboardViewer to
// facilitate receiving of clipboard change notifications. Could mix it in if
// an additional base was desired.
//
class TBmpViewWindow : virtual public TWindow, public TClipboardViewer {
  public:
    char      FileName[MAXPATH];
    TDib*     Dib;
#if defined(USE_DDB)
	 TBitmap*  Bitmap;
#endif
    TPalette* Palette;
	 TBrush*   BkgndBrush;
	 long      Rop;
    int       PixelWidth;
    int       PixelHeight;
	 WORD      Colors;
    BOOL      Fit;
    BOOL      AutoClipView;

    TBmpViewWindow();
   ~TBmpViewWindow();

  protected:
    void      CmFileOpen();
	 void      CmFit();
	 void      CmAutoClipView();
	 void      CeFit(TCommandEnabler& ce);
	 void      CeAutoClipView(TCommandEnabler& ce);

	 void      Paint(TDC&, BOOL erase, TRect&);
	 void      EvSize(UINT sizeType, TSize&);

	 void      EvPaletteChanged(HWND hWndPalChg);
	 BOOL      EvQueryNewPalette();
	 void      EvSetFocus(HWND); // should use above when working
	 void      EvDestroy();

	 BOOL      UpdatePalette(BOOL alwaysRepaint);
	 void      AdjustScroller();
	 void      SetCaption(const char*);
	 BOOL      LoadBitmapFile(const char*);

  DECLARE_RESPONSE_TABLE(TBmpViewWindow);
};

DEFINE_RESPONSE_TABLE2(TBmpViewWindow, TClipboardViewer, TWindow)
  EV_COMMAND(CM_FILEOPEN, CmFileOpen),
  EV_COMMAND(CM_FIT, CmFit),
  EV_COMMAND(CM_AUTOCLIPVIEW, CmAutoClipView),
  EV_COMMAND_ENABLE(CM_FIT, CeFit),
  EV_COMMAND_ENABLE(CM_AUTOCLIPVIEW, CeAutoClipView),
  EV_WM_SIZE,
  EV_WM_PALETTECHANGED,
  EV_WM_QUERYNEWPALETTE,
  EV_WM_SETFOCUS,
  EV_WM_DESTROY,
END_RESPONSE_TABLE;

//
// Constructor for a TBmpViewWindow, sets scroll styles and constructs
// the Scroller object.  Also sets the Rop based on whether the display
// is monochrome (two-color) or polychrome.
//
TBmpViewWindow::TBmpViewWindow()
  : TWindow(0, 0, 0)
{
  // Init TWindow since TClipboardViewer is usually a mixin and doesn't do it
  //
//  Init(0, 0, 0);

  Attr.Style |= WS_VSCROLL | WS_HSCROLL;
  Dib = 0;
#if defined(USE_DDB)
  Bitmap = 0;
#endif
  Palette = 0;
  BkgndBrush = new TBrush(::GetSysColor(COLOR_WINDOW));
  Scroller = new TScroller(this, 1, 1, 200, 200);
  Fit = FALSE;
  AutoClipView = FALSE;

  TScreenDC screenDC;
  if (screenDC.GetDeviceCaps(NUMCOLORS) < 3 )
    Rop = NOTSRCCOPY;
  else
    Rop = SRCCOPY;
  
  SetCaption(0);
}

TBmpViewWindow::~TBmpViewWindow()
{
#if defined(USE_DDB)
  delete Bitmap;
#endif
  delete Palette;
  delete Dib;
  delete BkgndBrush;
}

//
// Build up a caption based on a filename, and set it into the title.
//
void
TBmpViewWindow::SetCaption(const char* name) 
{
  char caption[MAXPATH + MAXAPPNAME + 2 + 1];
  strcpy(FileName, name ? name : "(Untitled)");
  strcpy(caption, GetApplication()->GetName());
  strcat(caption, " - ");
  strcat(caption, FileName);
  if (Parent)
	 Parent->SetCaption(caption);
}



//
// Toggle Fit member variable & adjust scroller as needed
//
void
TBmpViewWindow::CmFit()
{
  Fit = !Fit;
  AdjustScroller();
}

//
// The fit menu item is checked if the Fit member is true
//
void
TBmpViewWindow::CeFit(TCommandEnabler& ce)
{
  ce.SetCheck(Fit ? TCommandEnabler::Checked : TCommandEnabler::Unchecked);
}

//
// Toggle AutoAutoClipView member variable
//
void
TBmpViewWindow::CmAutoClipView()
{
  AutoClipView = !AutoClipView;
}

//
// Check AutoClipView according to flag
//
void
TBmpViewWindow::CeAutoClipView(TCommandEnabler& ce)
{
  ce.SetCheck(AutoClipView ? TCommandEnabler::Checked : TCommandEnabler::Unchecked);
}

//
// If the the "Open..." menu item is selected, then we prompt the user
// for a new bitmap file.  If the user selects one and it is one that
// we can read, we display it in the window and change the window's
// caption to reflect the new bitmap file.
//
void
TBmpViewWindow::CmFileOpen()
{
  static TOpenSaveDialog::TData data (
    OFN_HIDEREADONLY|OFN_FILEMUSTEXIST|OFN_NOREADONLYRETURN,
	 "JPEG Files (*.JPG)|*.jpg|",
	 0,
	 "",
	 "JPG"
  );
  if (TFileOpenDialog(this, data).Execute() == IDOK) {
	 char fileTitle[MAXPATH];
	 TOpenSaveDialog::GetFileTitle(data.FileName, fileTitle, MAXPATH);
	 LoadBitmapFile(fileTitle);
	 SetCaption(strlwr(fileTitle));
  }
}

//
// Adjust the Scroller range so that the the origin is the
// upper-most scrollable point and the corner is the
// bottom-most.
//
void
TBmpViewWindow::AdjustScroller() 
{
  TRect  clientRect = GetClientRect();

  // only show scrollbars when image is larger than
  // the client area and we are not stretching to fit.
  //
  if (Fit)
	 Scroller->SetRange(0, 0);

  else {
    TPoint Range(Max(PixelWidth-clientRect.Width(), 0),
					  Max(PixelHeight-clientRect.Height(), 0));
    Scroller->SetRange(Range.x, Range.y);
  }
  Scroller->ScrollTo(0, 0);
  if (!GetUpdateRect(clientRect, FALSE))
    Invalidate(FALSE);
}

//
// Reset scroller range.
//
void
TBmpViewWindow::EvSize(UINT SizeType, TSize& Size)
{
  TWindow::EvSize(SizeType, Size);
  if (SizeType != SIZEICONIC) {
	 AdjustScroller();
	 Invalidate(FALSE);
  }
}

//
// Somebody changed the palette. If its not us, then we need to update.
//
void
TBmpViewWindow::EvPaletteChanged(HWND hWndPalChg)
{
  if (hWndPalChg != HWindow)
	 UpdatePalette(TRUE);    // pass FALSE to UpdateColors() instead of repaint
}

//
// We need to re-realize the logical palette each time
// we regain the input focus
//
BOOL
TBmpViewWindow::EvQueryNewPalette()
{
  return UpdatePalette(TRUE);
}

//
// Use this message temporarily until the palette msgs get routed to us
//
void
TBmpViewWindow::EvSetFocus(HWND)
{
  UpdatePalette(TRUE);
}

void
TBmpViewWindow::EvDestroy()
{
  if (TClipboardViewer::DoDestroy() == esComplete)
    return;
  TWindow::EvDestroy();
}

BOOL
TBmpViewWindow::UpdatePalette(BOOL alwaysRepaint)
{
  TRACEX(BmpView, 2, "Enter UpdatePalette()");
  if (Palette) {
	 TClientDC clientDC(*this);
    #if !defined(__WIN32__)
      Palette->UnrealizeObject();
    #endif
    TRACEX(BmpView, 2, "UpdatePalette::Palette: " << hex << (UINT)(HPALETTE)*Palette);
    clientDC.SelectObject(*Palette, FALSE);
    if (clientDC.RealizePalette() > 0)
      if (alwaysRepaint)
	Invalidate(FALSE);
      else
	clientDC.UpdateColors();
	 TRACEX(BmpView, 2, "Leave UpdatePalette(TRUE)");
	 return TRUE;
  }
  TRACEX(BmpView, 2, "Leave UpdatePalette(FALSE)");
  return FALSE;
}

//
// Responds to an incoming Paint message by redrawing the bitmap.
// The Scroller's BeginView method, which sets the viewport origin
// relative to the present scroll position, has been called by TWindow's
// EvPaint
// Note that we Invalidate() ourselves with false to avoid the background 
// paint flicker. Thats why we use 
//
// The code can use either the Bitmap member with Stretch- and Bit- Blts,
// or use the Dib member with Stretch- and Set- DIBits... 
//
void
TBmpViewWindow::Paint(TDC& dc, BOOL, TRect&)
{
  TRACEX(BmpView, 2, "Enter Paint()");
  TRect clientRect = GetClientRect();
  TRACEX(BmpView, 2, "Paint::BkBrush: " << hex << (UINT)(HBRUSH)*BkgndBrush);
  dc.SelectObject(*BkgndBrush);

#if defined(USE_DDB)
  if (Bitmap) {
    TMemoryDC memoryDC(dc);
	 TRACEX(BmpView, 2, "Paint::Bitmap: " << hex << (UINT)(HBITMAP)*Bitmap);
    memoryDC.SelectObject(*Bitmap);
	 dc.SetStretchBltMode(COLORONCOLOR);
    if (Palette) {
      TRACEX(BmpView, 2, "Paint::Palette: " << hex << (UINT)(HPALETTE)*Palette);
      dc.SelectObject(*Palette, FALSE);
      dc.RealizePalette();
      memoryDC.SelectObject(*Palette, FALSE);
    }
    if (Rop == SRCCOPY) {
      if (Colors == 2) {
	if (Palette) {
	  PALETTEENTRY pe;
			 Palette->GetPaletteEntry(0, pe);
	  dc.SetTextColor(TColor(pe));
	  Palette->GetPaletteEntry(1, pe);
	  dc.SetBkColor(TColor(pe));
	} else {
	  dc.SetBkColor(TColor(0,0,0));
			 dc.SetTextColor(TColor(255,255,255));
	}
		}
    }
#else
  if (Dib) {
#endif

    TRect imageRect(0,0, PixelWidth, PixelHeight);
    if (Parent->IsIconic()) {
      #if defined(USE_DDB)
	dc.StretchBlt(Parent->GetClientRect(), memoryDC, imageRect, Rop);
      #else
	dc.StretchDIBits(Parent->GetClientRect(), imageRect, *Dib, Rop);
		#endif

    } else {
      clientRect += TPoint((int)Scroller->XPos, (int)Scroller->YPos);
      if (Fit) {
	#if defined(USE_DDB)
	  dc.StretchBlt(clientRect, memoryDC, imageRect, Rop);
		  #else
	  dc.StretchDIBits(clientRect, imageRect, *Dib);
		  #endif

		} else {
	#if defined(USE_DDB)
	  dc.BitBlt(imageRect, memoryDC, TPoint(0,0), Rop);
	#else
	  dc.SetDIBitsToDevice(imageRect, TPoint(0,0), *Dib);
	#endif

	// Clear borders here for no flicker
		  //
	dc.PatBlt(TRect(TPoint(PixelWidth,0), clientRect.BottomRight()));
	dc.PatBlt(TRect(TPoint(0,PixelHeight), clientRect.BottomRight()));
		}
	 }
  } else
	 dc.PatBlt(clientRect, PATCOPY);
  TRACEX(BmpView, 2, "Leave Paint()" << endl);
}

//
// Test if the passed file is a Windows 3.0 DI (or PM 1.x) bitmap
// and if so read it.
// Report errors if unable to do so. Adjust the Scroller to the new
// bitmap dimensions.
//
BOOL
TBmpViewWindow::LoadBitmapFile(const char* name)
{
  int MaxX, MaxY;
  int pad, i;
  struct jpeg_decompress_struct cinfo;
  struct jpeg_error_mgr jerr;
  FILE *fpin;
  JSAMPARRAY buffer;
  int row_stride;
  if ((fpin = fopen(name, "rb")) == NULL)
	{
	 MessageBox(name, "Error Opening File", MB_OK);
	 return(FALSE);
	}
  cinfo.err = jpeg_std_error(&jerr);
  jpeg_create_decompress(&cinfo);
  jpeg_stdio_src(&cinfo, fpin);
  (void) jpeg_read_header(&cinfo, TRUE);

  MaxX = cinfo.image_width;
  MaxY = cinfo.image_height;
  int remain = MaxX % 4;
  if (remain > 0)
	{
	 pad = (4 - remain);
	 MaxX = MaxX + (4 - remain);
	}
  remain = MaxY % 2;
  if (remain > 0)
	 MaxY = MaxY + 1;

  delete Dib;
  Dib = new TDib(MaxX, MaxY, 256, DIB_RGB_COLORS);

  cinfo.quantize_colors = TRUE;
  jpeg_start_decompress(&cinfo);

  // do colortable entries

  JSAMPARRAY colormap = cinfo.colormap;
  int num_colors = cinfo.actual_number_of_colors;
  for (i = 0; i < num_colors; i++)
	{
	 Dib->SetColor(i, TColor((GETJSAMPLE(colormap[0][i])),
			  (GETJSAMPLE(colormap[1][i])), (GETJSAMPLE(colormap[2][i]))));
	}
  delete Palette;
  Palette = new TPalette(*Dib);

  // read in data

  row_stride = cinfo.output_width * cinfo.output_components;
  buffer = (*cinfo.mem->alloc_sarray)
			  ((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);
  unsigned char huge *pcntr;
  void huge *pdib;
  pdib = Dib->GetBits();
  unsigned char huge *pincr = (unsigned char huge *)pdib;
  unsigned char far *pbuf;
  pcntr = pincr;
  long j;
  for (j = 0; j < (MaxX * MaxY); j++)
	{
	 *pcntr = 0;
	 pcntr++;
	}
  for (j = 0; j < cinfo.output_height; j++)
	{
	 (void) jpeg_read_scanlines(&cinfo, buffer, 1);
	 pcntr = pincr;
	 pcntr = pcntr + ((cinfo.output_height - j - 1)*MaxX);
	 pbuf = buffer[0];
	 for (i = 0; i < (MaxX - pad); i++)
	  {
		*pcntr = *pbuf;
		++pcntr;
		++pbuf;
	  }
	 for (i = 0; i < pad; i++)
	  {
		*pcntr = 0;
		++pcntr;
	  }
	}


  jpeg_finish_decompress(&cinfo);
  jpeg_destroy_decompress(&cinfo);
  fclose(fpin);

#if defined(USE_DDB)
  delete Bitmap;
  Bitmap = new TBitmap(*Dib, Palette);
#endif
  PixelWidth  = Dib->Width();
  PixelHeight = Dib->Height();
  UpdatePalette(TRUE);
  AdjustScroller();
#if defined(PAINT_ICON)
  Parent->SetIcon(0, 0);  // enable to paint the icon from the bitmap
#endif
  return TRUE;
}



//----------------------------------------------------------------------------

class TBmpViewApp : public TApplication {
  public:
	 TBmpViewApp(const char far* name) : TApplication(name) {}
    void InitMainWindow() {
      MainWindow = new TFrameWindow(0, Name, new TBmpViewWindow);
      MainWindow->AssignMenu(Name);
      MainWindow->SetIcon(this, Name);
    }
};

//----------------------------------------------------------------------------

int
OwlMain(int /*argc*/, char* /*argv*/ [])
{
  return TBmpViewApp(AppName).Run();
}
